/**
 * Copyright (C) 2012  Dušan Vejnovič  <vaadin@dussan.org>
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 * http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */
package org.dussan.vaadin.dtable.client.ui;

import static com.google.gwt.query.client.GQuery.$;
import static com.google.gwt.query.client.GQuery.window;

import java.util.ArrayList;
import java.util.Collections;
import java.util.HashMap;
import java.util.List;
import java.util.Map;

import org.dussan.vaadin.dtable.client.TableConfig;
import org.dussan.vaadin.dtable.client.TableExporter;
import org.dussan.vaadin.dtable.client.Translation;
import org.dussan.vaadin.dtable.client.components.Filtering;
import org.dussan.vaadin.dtable.client.components.LoadingMessage;
import org.dussan.vaadin.dtable.client.components.Pager;
import org.dussan.vaadin.dtable.client.components.QuickSearch;
import org.dussan.vaadin.dtable.client.components.SearchingColumns;
import org.dussan.vaadin.dtable.client.components.Sorting;
import org.dussan.vaadin.dtable.client.components.Splitter;
import org.dussan.vaadin.dtable.client.components.TableTools;
import org.dussan.vaadin.dtable.client.components.events.TableEvent;
import org.dussan.vaadin.dtable.client.components.menu.Menu;
import org.dussan.vaadin.dtable.client.components.table.TableBody;
import org.dussan.vaadin.dtable.client.components.table.TableHeader;

import com.google.gwt.dom.client.Style.Cursor;
import com.google.gwt.dom.client.Style.Display;
import com.google.gwt.dom.client.Style.Unit;
import com.google.gwt.query.client.Function;
import com.google.gwt.query.client.GQuery;
import com.google.gwt.query.client.css.CSS;
import com.google.gwt.query.client.css.Length;
import com.google.gwt.query.client.css.TakesCssValue.CssSetter;
import com.google.gwt.query.client.css.TextAlignProperty.TextAlign;
import com.google.gwt.user.client.DOM;
import com.google.gwt.user.client.Element;
import com.google.gwt.user.client.Event;
import com.google.gwt.user.client.Timer;
import com.google.gwt.user.client.Window;
import com.google.gwt.user.client.ui.SimplePanel;
import com.vaadin.terminal.gwt.client.ApplicationConnection;
import com.vaadin.terminal.gwt.client.BrowserInfo;
import com.vaadin.terminal.gwt.client.Paintable;
import com.vaadin.terminal.gwt.client.UIDL;

public class VDTable extends SimplePanel implements Paintable {

	public static final TextAlign[] COLUMN_ALIGN = { TextAlign.LEFT,
			TextAlign.RIGHT, TextAlign.CENTER };

	private Boolean fetchNewData = null; // don't change
	private boolean allColumnsVisible = false;
	private boolean allRowsVisible = false;
	private boolean enableResizeTimer = false;
	private boolean enableScrollTimer = false;
	private boolean hideFrozenTable = false;
	private boolean showFrozenTable = false;
	private boolean sendingMessage = false;
	private boolean isAnyFrozenColumnHidden = false;

	private int id = 0;
	private int left = 0;
	private int top = 0;
	private int browserOffset = 0;
	private int tableFrozenWidth = 0;
	private int tableWidth = 0;
	private int firstRow = 0;
	private int allColumns = 0;
	private int allRows = 0;
	private int windowX = 0;
	private int windowY = 0;
	private int scrollLeftFrozen = 0;
	private int scrollLeft = 0;
	private int scrollTop = 0;
	private int scrollTopFirstVisibleRow = 0;
	private int controledScroll = 0;
	private int ffMarginTop = 0;

	protected String uidl = null;
	protected ApplicationConnection client = null;

	private Translation translation = null;
	private TableConfig tableConfig = null;
	private TableExporter tableExporter = null;

	private TableHeader tableHeader = null;
	private TableBody tableBody = null;
	private Splitter splitter = null;
	private TableTools tableTools = null;
	private QuickSearch quickSearch = null;
	private SearchingColumns searchColumns = null;
	private Sorting sorting = null;
	private Filtering filtering = null;
	private LoadingMessage loadingMessage = null;
	private Menu menu = null;
	private Pager pager = null;

	private Map<String, Object> request = new HashMap<String, Object>();
	private Map<Integer, Integer> columnsWidth = new HashMap<Integer, Integer>();
	private List<Integer> columnsFrozen = new ArrayList<Integer>();
	private List<Integer> columnsVisible = new ArrayList<Integer>();

	private GQuery mouseLeaveWindowEventHandler = null;
	private GQuery windowResizeEventHandler = null;

	private GQuery tableOutEventHandler = null;
	private GQuery tableScrollEventHandler = null;
	private GQuery frozenTableScrollEventHandler = null;

	public VDTable() {
		// global id
		id = (int) (Math.random() * 1000000);
		getElement().setId("v-dtable" + id);
		setStyleName("v-dtable");

		translation = new Translation();
		tableConfig = new TableConfig();
		tableExporter = new TableExporter();

		tableHeader = new TableHeader(this);
		tableBody = new TableBody(this);
		splitter = new Splitter(this);
		tableTools = new TableTools(this);
		quickSearch = new QuickSearch(this);
		searchColumns = new SearchingColumns(this);
		sorting = new Sorting(this);
		filtering = new Filtering(this);
		loadingMessage = new LoadingMessage(this);
		pager = new Pager(this);
	}

	private void initContainers() {
		browserOffset = BrowserInfo.get().isChrome() ? 0 : 1;

		Element mainContainerWrapper = DOM.createDiv();
		mainContainerWrapper.setId("dt-main-container-wrapper");
		getElement().appendChild(mainContainerWrapper);

		Element mainContainer = DOM.createTable();
		mainContainer.setId("dt-main-container");
		mainContainerWrapper.appendChild(mainContainer);

		Element message = DOM.createDiv();
		message.setId("dt-message" + getId());
		message.setClassName("dt-message");
		$(message).css(CSS.DISPLAY.with(Display.NONE));
		mainContainerWrapper.appendChild(message);

		// header section
		Element trHeader = DOM.createTR();
		mainContainer.appendChild(trHeader);

		Element headerWrapper = DOM.createTD();
		headerWrapper.setId("dt-header-wrapper");
		trHeader.appendChild(headerWrapper);

		Element headerContainer = DOM.createDiv();
		headerContainer.setId("dt-header-container");
		headerWrapper.appendChild(headerContainer);

		Element headerTable = DOM.createTable();
		headerTable.setId("dt-header-table" + getId());
		headerTable.getStyle().setWidth(100, Unit.PCT);
		headerContainer.appendChild(headerTable);

		Element headerContainerLength = DOM.createDiv();
		headerContainerLength.setId("dt-header-container-length" + getId());
		headerContainerLength.setClassName("dt-header-container-length");
		headerContainer.appendChild(headerContainerLength);

		Element headerWrapperEmpty = DOM.createTD();
		headerWrapperEmpty.setId("dt-header-wrapper-empty");
		trHeader.appendChild(headerWrapperEmpty);

		// body section
		Element trBody = DOM.createTR();
		mainContainer.appendChild(trBody);

		Element bodyWrapper = DOM.createTD();
		bodyWrapper.setId("dt-body-wrapper");
		bodyWrapper.setPropertyInt("colSpan", 2);
		trBody.appendChild(bodyWrapper);

		Element bodyContainer = DOM.createDiv();
		bodyContainer.setId("dt-body-container");
		bodyWrapper.appendChild(bodyContainer);

		Element bodyTable = DOM.createTable();
		bodyTable.setId("dt-body-table" + getId());
		bodyTable.getStyle().setWidth(100, Unit.PCT);
		bodyContainer.appendChild(bodyTable);

		Element bodyContainerLength = DOM.createDiv();
		bodyContainerLength.setId("dt-body-container-length" + getId());
		bodyContainerLength.setClassName("dt-body-container-length");
		bodyContainer.appendChild(bodyContainerLength);

		if (menu == null) {
			menu = new Menu(this);
			menu.setGlassEnabled(true);
		}
	}

	private void appendFrozenTable() {
		// header section
		Element frozenHeaderWrapper = DOM.createTD();
		frozenHeaderWrapper.setId("dt-frozen-header-wrapper");

		Element frozenHeaderContainer = DOM.createDiv();
		frozenHeaderContainer.setId("dt-frozen-header-container");
		frozenHeaderWrapper.appendChild(frozenHeaderContainer);

		Element frozenHeaderTable = DOM.createTable();
		frozenHeaderTable.setId("dt-frozen-header-table" + getId());
		frozenHeaderTable.getStyle().setWidth(100, Unit.PCT);
		frozenHeaderContainer.appendChild(frozenHeaderTable);

		Element splitterContainer = DOM.createTD();
		splitterContainer.setId("dt-splitter-container");
		splitterContainer.setAttribute("rowspan", "2");
		splitterContainer.getStyle().setCursor(Cursor.DEFAULT);

		$("table#dt-header-table" + getId()).parent().parent()
				.before($(frozenHeaderWrapper));
		$("table#dt-header-table" + getId()).parent().parent()
				.before($(splitterContainer));

		// body section
		Element frozenBodyWrapper = DOM.createTD();
		frozenBodyWrapper.setId("dt-frozen-body-wrapper");

		Element frozenBodyContainer = DOM.createDiv();
		frozenBodyContainer.setId("dt-frozen-body-container");
		frozenBodyWrapper.appendChild(frozenBodyContainer);

		Element frozenBodyTable = DOM.createTable();
		frozenBodyTable.setId("dt-frozen-body-table" + getId());
		frozenBodyTable.getStyle().setWidth(100, Unit.PCT);
		frozenBodyContainer.appendChild(frozenBodyTable);

		Element frozenBodyContainerLength = DOM.createDiv();
		frozenBodyContainerLength.setId("dt-frozen-body-container-length"
				+ getId());
		frozenBodyContainerLength
				.setClassName("dt-frozen-body-container-length");
		frozenBodyContainer.appendChild(frozenBodyContainerLength);

		$("table#dt-body-table" + getId()).parent().parent()
				.before($(frozenBodyWrapper));
	}

	private void loadComponents() {
		tableHeader.load();
		tableBody.load();
		splitter.load();
		tableTools.load();
		quickSearch.load();
		searchColumns.load();
		sorting.load();
		filtering.load();
		loadingMessage.load();
		pager.load();
	}

	private void unloadComponents() {
		tableHeader.unload();
		tableBody.unload();
		splitter.unload();
		tableTools.unload();
		quickSearch.unload();
		searchColumns.unload();
		sorting.unload();
		filtering.unload();
		loadingMessage.unload();
		pager.unload();
	}

	private void activateWindowResizeEventHandler(boolean enable) {
		if (enable) {
			activateWindowResizeEventHandler(false);
			windowResizeEventHandler = $(window).resize(new Function() {
				public void f() {
					$("div#dt-header-container-length" + getId()).css(
							CSS.WIDTH.with(Length.px(tableWidth)));
					$("div#dt-body-container-length" + getId()).css(
							CSS.WIDTH.with(Length.px(tableWidth)));
					if (!enableResizeTimer) {
						updateTableOnResizing();
					}
				};
			});
		} else if (!enable && windowResizeEventHandler != null) {
			windowResizeEventHandler.die();
			windowResizeEventHandler = null;
		}
	}

	private void activateMouseLeaveWindowEventHandler(boolean enable) {
		if (enable) {
			activateMouseLeaveWindowEventHandler(false);
			mouseLeaveWindowEventHandler = $(window).mouseout(new Function() {
				public boolean f(Event e) {
					TableEvent.mouseLeaveWindow(
							$("table#dt-frozen-header-table" + getId()),
							$("table#dt-frozen-header-table" + getId()),
							$("table#dt-frozen-body-table" + getId()),
							$("table#dt-body-table" + getId()), config()
									.isEnabledHorizontalLines());
					return false;
				}
			});
		} else if (!enable && mouseLeaveWindowEventHandler != null) {
			mouseLeaveWindowEventHandler.die();
			mouseLeaveWindowEventHandler = null;
		}
	}

	private void activateTableOutEventHandler(boolean enable) {
		if (enable) {
			activateTableOutEventHandler(false);
			tableOutEventHandler = $("div#v-dtable" + getId()).mouseout(
					new Function() {
						public void f() {
							TableEvent
									.mouseLeaveWindow(
											$("table#dt-frozen-header-table"
													+ getId()),
											$("table#dt-frozen-header-table"
													+ getId()),
											$("table#dt-frozen-body-table"
													+ getId()),
											$("table#dt-body-table" + getId()),
											config().isEnabledHorizontalLines());
						};
					});
		} else if (!enable && tableOutEventHandler != null) {
			tableOutEventHandler.die();
			tableOutEventHandler = null;
		}
	}

	private void activateTableScrollEventHandler(boolean enable) {
		if (enable) {
			final GQuery header = $("table#dt-header-table" + getId()).parent();
			final GQuery body = $("table#dt-body-table" + getId()).parent();
			final GQuery frozenBody = $("table#dt-frozen-body-table" + getId());

			activateTableScrollEventHandler(false);
			tableScrollEventHandler = body.scroll(new Function() {
				public void f() {
					header.scrollLeft(body.scrollLeft());
					if (BrowserInfo.get().isFirefox()) {
						// FF bug: flickering if vertical scroll bar is hidden
						frozenBody.css(CSS.MARGIN_TOP.with(Length.px(-body
								.scrollTop() + ffMarginTop)));
					} else {
						frozenBody.parent().scrollTop(body.scrollTop());
					}

					if (!enableScrollTimer) {
						updateTableOnScrolling();
					}
				};
			});
		} else if (!enable && tableScrollEventHandler != null) {
			tableScrollEventHandler.die();
			tableScrollEventHandler = null;
		}
	}

	private void activateFrozenTableScrollEventHandler(boolean enable) {
		if (enable) {
			final GQuery header = $("table#dt-frozen-header-table" + getId())
					.parent();
			final GQuery body = $("table#dt-frozen-body-table" + getId())
					.parent();
			activateFrozenTableScrollEventHandler(false);
			frozenTableScrollEventHandler = body.scroll(new Function() {
				public void f() {
					scrollLeftFrozen = body.scrollLeft();
					header.scrollLeft(scrollLeftFrozen);
				};
			});
		} else if (!enable && frozenTableScrollEventHandler != null) {
			frozenTableScrollEventHandler.die();
			frozenTableScrollEventHandler = null;
		}
	}

	private void updateTableOnResizing() {
		if (windowX != Window.getClientWidth()
				|| windowY != Window.getClientHeight()) {
			enableResizeTimer = true;

			Timer timer = new Timer() {
				public void run() {
					int winX = Window.getClientWidth();
					int winY = Window.getClientHeight();
					GQuery frozenBody = $("table#dt-frozen-body-table"
							+ getId());
					GQuery body = $("table#dt-body-table" + getId());

					if (windowX == winX && windowY == winY) {
						// check if table width is smaller then
						// table default minimal width
						if (config().isEnabledFrozenTable()
								&& isFrozenTableVisible()) {
							if (body.parent().outerWidth() < config()
									.getTableMinimalWidth()) {
								String[][] filter = {
										{ "td#dt-frozen-header-wrapper",
												"td#dt-frozen-body-wrapper" },
										{ "div#dt-frozen-header-container",
												"div#dt-frozen-body-container" } };
								GQuery mainContainerWrapper = $(
										"div#v-dtable" + getId()).find(
										"div#dt-main-container-wrapper");
								mainContainerWrapper
										.children("table")
										.children("tr")
										.children(filter[0])
										.css(CSS.WIDTH.with(Length.px(config()
												.getTableFrozenMinimalWidth())));
								mainContainerWrapper
										.children("table")
										.children("tr")
										.children("td")
										.children(filter[1])
										.css(CSS.WIDTH.with(Length.px(config()
												.getTableFrozenMinimalWidth())));
							}
						}

						cancel();
						enableResizeTimer = false;
						checkContainerDimensionsAndTableCellAlignments(
								frozenBody, body);
						forceUpdateClientLayout();
					}

					windowX = Window.getClientWidth();
					windowY = Window.getClientHeight();
				}
			};
			timer.scheduleRepeating(config().getPageResizingActionDelay());
		}
	}

	private void updateTableOnScrolling() {
		final GQuery body = $("table#dt-body-table" + getId()).parent();
		if (scrollLeft != body.scrollLeft() || scrollTop != body.scrollTop()) {
			enableScrollTimer = true;

			Timer timer = new Timer() {
				public void run() {
					int left = body.scrollLeft();
					int top = body.scrollTop();

					if (scrollLeft == left && scrollTop == top) {
						if (!config().isEnabledAllRows()
								|| !config().isEnabledAllColumns()) {
							fetchNewData = scrollLeft == controledScroll;
							controledScroll = scrollLeft;
							if ((fetchNewData && !allRowsVisible)
									|| (!fetchNewData && !allColumnsVisible)) {
								sendMessage("newData", fetchNewData.toString(),
										true);
							}
						}

						cancel();
						enableScrollTimer = false;
						fetchNewData = null;
					}

					scrollLeft = left;
					scrollTop = top;
				}
			};
			timer.scheduleRepeating(config().getScrollbarActionDelay());
		}
	}

	public void forceUpdateClientLayout() {
		client.forceLayout();
	}

	public void enableEventsHandler() {
		tableHeader.enableEventsHandler();
		tableBody.enableEventsHandler();
		splitter.enableEventsHandler();
		tableTools.enableEventsHandler();
		quickSearch.enableEventsHandler();
		searchColumns.enableEventsHandler();
		sorting.enableEventsHandler();
		filtering.enableEventsHandler();
		loadingMessage.enableEventsHandler();
		pager.enableEventsHandler();
		sendingMessage = false;
	}

	public void disableEventsHandler() {
		sendingMessage = true;
		tableHeader.disableEventsHandler();
		tableBody.disableEventsHandler();
		splitter.disableEventsHandler();
		tableTools.disableEventsHandler();
		quickSearch.disableEventsHandler();
		searchColumns.disableEventsHandler();
		sorting.disableEventsHandler();
		filtering.disableEventsHandler();
		loadingMessage.disableEventsHandler();
		pager.disableEventsHandler();
	}

	public Translation translation() {
		return translation;
	}

	public TableConfig config() {
		return tableConfig;
	}

	public TableHeader header() {
		return tableHeader;
	}

	public TableExporter exporter() {
		return tableExporter;
	}

	public TableTools tools() {
		return tableTools;
	}

	public QuickSearch quickSearch() {
		return quickSearch;
	}

	public SearchingColumns searchColumns() {
		return searchColumns;
	}

	public Sorting sorting() {
		return sorting;
	}

	public Filtering filtering() {
		return filtering;
	}

	public Menu menu() {
		return menu;
	}

	public int getId() {
		return id;
	}

	public int getAllColumns() {
		return allColumns;
	}

	public int getFirstRow() {
		return firstRow;
	}

	public int getAllRows() {
		return allRows;
	}

	public int getFrozenTableWidth() {
		return tableFrozenWidth;
	}

	public void setFrozenTableWidth(int tableFrozenWidth) {
		this.tableFrozenWidth = tableFrozenWidth;
	}

	public int getLeft() {
		return left;
	}

	public int getScrollLeftFrozen() {
		return scrollLeftFrozen;
	}

	public int getScrollLeft() {
		return scrollLeft;
	}

	public int getBrowserOffset() {
		return browserOffset;
	}

	public List<Integer> getColumnsFrozen() {
		return columnsFrozen;
	}

	public List<Integer> getColumnsVisible() {
		return columnsVisible;
	}

	public Map<Integer, Integer> getColumnsWidth() {
		return columnsWidth;
	}

	public boolean isFrozenTableHidden() {
		return hideFrozenTable;
	}

	public boolean isFrozenTableVisible() {
		return (showFrozenTable && !hideFrozenTable);
	}

	public boolean isAllColumnsVisible() {
		return allColumnsVisible;
	}

	public boolean isAnyFrozenColumnHidden() {
		return isAnyFrozenColumnHidden;
	}

	public void setAllColumnsVisible(boolean allColumnsVisible) {
		this.allColumnsVisible = allColumnsVisible;
	}

	public void showFrozenTable(boolean show) {
		final String[][] filter = {
				{ "td#dt-frozen-header-wrapper", "td#dt-frozen-body-wrapper" },
				{ "div#dt-frozen-header-container",
						"div#dt-frozen-body-container" } };
		GQuery mainContainer = $("div#v-dtable" + getId()
				+ " table#dt-main-container");

		if (show && !showFrozenTable && !hideFrozenTable) {
			showFrozenTable = show;
			mainContainer.children("tr").children(filter[0])
					.css(CSS.WIDTH.with(Length.px(tableFrozenWidth)));
			mainContainer.children("tr").children("td").children(filter[1])
					.css(CSS.WIDTH.with(Length.px(tableFrozenWidth)));

			// set width to 100% - Chrome want this
			mainContainer.children("tr").children("td").children(filter[1])
					.css(CSS.WIDTH.with(Length.pct(100)));
		} else if (!show && showFrozenTable) {
			showFrozenTable = hideFrozenTable;
			mainContainer.children("tr").children(filter[0])
					.css(CSS.WIDTH.with(Length.px(0)));
			mainContainer.children("tr").children("td").children(filter[1])
					.css(CSS.WIDTH.with(Length.px(0)));
		}
	}

	public boolean isFrozen(int column) {
		return columnsFrozen.contains(column);
	}

	public void setFrozen(int column) {
		// remove filtering
		if ((!config().isEnabledFilteredFrozenTable() && !isFrozen(column) && filtering
				.isFiltered(column))
				|| (!config().isEnabledFilteredTable() && isFrozen(column) && filtering
						.isFiltered(column))) {
			request.put("removeFilter", Integer.toString(column));
		}

		// remove sorting
		if (!config().isEnabledSortedFrozenTable() && !isFrozen(column)) {
			$("table#dt-header-table" + getId()).children("tr")
					.children("td [id='" + getId() + ":" + column + "']")
					.removeClass("asc", "desc");
			request.put("sorting", sorting.getSortings());
		} else if (!config().isEnabledSortedTable() && isFrozen(column)) {
			$("table#dt-frozen-header-table" + getId()).children("tr")
					.children("td [id='" + getId() + ":" + column + "']")
					.removeClass("asc", "desc");
			request.put("sorting", sorting.getSortings());
		}

		sendMessage("freeze", Integer.toString(column), true);
	}

	private void setContainerWidthAndHeight(GQuery table, GQuery virtualTable,
			CssSetter tableHeight, CssSetter virtualTableWidth,
			CssSetter virtualTableHeight) {
		table.css(tableHeight);
		virtualTable.css(virtualTableWidth).css(virtualTableHeight);
	}

	private void setScrollableContainerWidthAndHeight(
			GQuery virtualHeaderTable, GQuery table, GQuery virtualTable,
			GQuery frozenTable, GQuery frozenVirtualTable, Length tableHeight,
			Length virtualTableWidth, Length virtualTableHeight,
			Length frozenVirtualTableHeight) {
		// set header container width and height
		virtualHeaderTable.css(CSS.WIDTH.with(virtualTableWidth))
				.css(CSS.HEIGHT.with(Length.px(virtualHeaderTable.parent()
						.height())));

		// set frozen body container width and height
		if (config().isEnabledFrozenTable()) {
			setContainerWidthAndHeight(frozenTable, frozenVirtualTable,
					CSS.HEIGHT.with(tableHeight),
					CSS.WIDTH.with(Length.pct(100)),
					CSS.HEIGHT.with(frozenVirtualTableHeight));
		}

		// set body container width and height
		setContainerWidthAndHeight(table, virtualTable,
				CSS.HEIGHT.with(tableHeight),
				CSS.WIDTH.with(virtualTableWidth),
				CSS.HEIGHT.with(virtualTableHeight));
	}

	private void setPageableContainerWidthAndHeight(GQuery virtualHeaderTable,
			GQuery table, GQuery virtualTable, GQuery frozenTable,
			GQuery frozenVirtualTable, Length tableHeight,
			Length virtualTableWidth) {
		// set header container width and height
		virtualHeaderTable.css(CSS.WIDTH.with(virtualTableWidth))
				.css(CSS.HEIGHT.with(Length.px(virtualHeaderTable.parent()
						.height())));

		// set frozen body container width and height
		if (config().isEnabledFrozenTable()) {
			setContainerWidthAndHeight(frozenTable, frozenVirtualTable,
					CSS.HEIGHT.with(tableHeight),
					CSS.WIDTH.with(Length.pct(100)),
					CSS.HEIGHT.with(Length.pct(100)));
		}

		// set body container width and height
		setContainerWidthAndHeight(table, virtualTable,
				CSS.HEIGHT.with(tableHeight),
				CSS.WIDTH.with(virtualTableWidth),
				CSS.HEIGHT.with(Length.pct(100)));
	}

	private void setNonfixedContainerWidthAndHeight(GQuery virtualHeaderTable,
			GQuery table, GQuery virtualTable, GQuery frozenTable,
			GQuery frozenVirtualTable, Length virtualTableWidth) {
		// set header container width and height
		virtualHeaderTable.css(CSS.WIDTH.with(virtualTableWidth))
				.css(CSS.HEIGHT.with(Length.px(virtualHeaderTable.parent()
						.height())));

		// set frozen body container width and height
		if (config().isEnabledFrozenTable()) {
			setContainerWidthAndHeight(frozenTable, frozenVirtualTable,
					CSS.HEIGHT.with(Length.pct(100)),
					CSS.WIDTH.with(Length.pct(100)),
					CSS.HEIGHT.with(Length.pct(100)));
		}

		// set body container width and height
		setContainerWidthAndHeight(table, virtualTable,
				CSS.HEIGHT.with(Length.pct(100)),
				CSS.WIDTH.with(virtualTableWidth),
				CSS.HEIGHT.with(Length.pct(100)));
	}

	private void setContainerWidthAndHeight() {
		// calculate horizontal scrollbar height
		GQuery body = $("table#dt-body-table" + getId()).parent();
		int tableHeight = allRows * config().getBodyRowHeight();
		int frozenScrollbarHeight = 0;
		int scrollbarHeight = 0;
		if (config().isEnabledFrozenTable() && isFrozenTableVisible()) {
			GQuery frozenBody = $("table#dt-frozen-body-table" + getId())
					.parent();
			frozenScrollbarHeight = frozenBody.outerHeight()
					- frozenBody.height();
		}
		scrollbarHeight = body.outerHeight() - body.height();

		if (!config().isEnabledPageableTable()
				&& (config().isEnabledFixedTableHeight() || config()
						.getTableHeight() <= tableHeight)) {
			// calculate scrollbar's addition to body container height
			int frozenTableHeightOffset = Math.max(0, scrollbarHeight
					- frozenScrollbarHeight);
			int tableHeightOffset = Math.max(0, frozenScrollbarHeight
					- scrollbarHeight);
			frozenTableHeightOffset += frozenTableHeightOffset > 0 ? browserOffset
					: 0;
			tableHeightOffset += tableHeightOffset > 0 ? browserOffset : 0;

			setScrollableContainerWidthAndHeight(
					$("div#dt-header-container-length" + getId()),
					$("table#dt-body-table" + getId()).parent(),
					$("div#dt-body-container-length" + getId()),
					$("table#dt-frozen-body-table" + getId()).parent(),
					$("div#dt-frozen-body-container-length" + getId()),
					Length.px(config().getTableHeight()), // tableHeight
					allColumnsVisible ? Length.pct(100) : Length.px(tableWidth), // virtualTableWidth
					Length.px(config().getBodyRowHeight() * allRows
							+ tableHeightOffset), // virtualTableHeight
					Length.px(config().getBodyRowHeight() * allRows
							+ frozenTableHeightOffset)); // frozenVirtualTableHeight
		} else if (config().isEnabledPageableTable()
				&& config().isEnabledFixedTableHeight()
				&& pager.getAllPages() > 1) {
			setPageableContainerWidthAndHeight(
					$("div#dt-header-container-length" + getId()),
					$("table#dt-body-table" + getId()).parent(),
					$("div#dt-body-container-length" + getId()),
					$("table#dt-frozen-body-table" + getId()).parent(),
					$("div#dt-frozen-body-container-length" + getId()),
					Length.px(config().getTableHeight()
							+ Math.max(frozenScrollbarHeight, scrollbarHeight)
							+ 1), // virtualTableHeight
					allColumnsVisible ? Length.pct(100) : Length.px(tableWidth)); // virtualTableWidth
		} else {
			setNonfixedContainerWidthAndHeight(
					$("div#dt-header-container-length" + getId()),
					$("table#dt-body-table" + getId()).parent(),
					$("div#dt-body-container-length" + getId()),
					$("table#dt-frozen-body-table" + getId()).parent(),
					$("div#dt-frozen-body-container-length" + getId()),
					allColumnsVisible ? Length.pct(100) : Length.px(tableWidth)); // virtualTableWidth
		}

		// add vertical scrollbar's width to empty header wrapper
		$("div#v-dtable" + getId() + " td#dt-header-wrapper-empty")
				.css(CSS.WIDTH.with(Length.px(body.outerWidth() - body.width()
						- 1)));
	}

	public void checkTableScrollers(GQuery frozenHeader, GQuery frozenBody,
			GQuery header, GQuery body) {
		// set body container width and height
		setContainerWidthAndHeight();

		scrollTopFirstVisibleRow = -scrollTop + top;
		if (config().isEnabledFrozenTable()) {
			frozenHeader.parent().scrollLeft(scrollLeftFrozen);
			frozenBody.parent().scrollLeft(scrollLeftFrozen);
			if (BrowserInfo.get().isFirefox()) {
				// FF bug: flickering if vertical scrollbar is hidden
				frozenBody.css(CSS.MARGIN_TOP.with(Length
						.px(scrollTopFirstVisibleRow)));
				ffMarginTop = top;
			} else {
				// set body container position
				frozenBody.css(CSS.TOP.with(Length.px(top)));
				// scroll table body into origin position
				frozenBody.parent().scrollTop(scrollTop);
			}
		}

		// set header container position
		header.css(CSS.LEFT.with(Length.px(!config()
				.isEnabledAllHeaderColumns() ? left : 0)));
		header.parent().scrollLeft(scrollLeft);

		// set body container position
		body.css(CSS.LEFT.with(Length.px(left)));
		body.css(CSS.TOP.with(Length.px(top)));

		// scroll table body into origin position
		body.parent().scrollLeft(scrollLeft);
		body.parent().scrollTop(scrollTop);
	}

	public void checkContainerDimensionsAndTableCellAlignments(
			GQuery frozenBody, GQuery body) {
		setContainerWidthAndHeight();
		tableBody.checkTdHeightsAndAlignments(frozenBody.children("tr"), true);
		tableBody.checkTdHeightsAndAlignments(body.children("tr"), true);
	}

	private int prepareTable(UIDL uidl) {
		List<Integer> frozenAndVisibleColumns = new ArrayList<Integer>();
		columnsFrozen.clear();
		if (uidl.hasAttribute("frozenColumns")) {
			int[] frozenColumns = uidl.getIntArrayAttribute("frozenColumns");
			for (int column : frozenColumns) {
				columnsFrozen.add(column);
			}
			frozenAndVisibleColumns.addAll(columnsFrozen);
		}
		columnsVisible.clear();
		if (uidl.hasAttribute("visibleColumns")) {
			int[] visibleColumns = uidl.getIntArrayAttribute("visibleColumns");
			for (int column : visibleColumns) {
				columnsVisible.add(column);
			}
			frozenAndVisibleColumns.addAll(columnsVisible);
		}
		Collections.sort(frozenAndVisibleColumns);

		if (uidl.hasAttribute("columnsWidth")) {
			columnsWidth.clear();
			int[] widths = uidl.getIntArrayAttribute("columnsWidth");
			for (int column = 0; column < frozenAndVisibleColumns.size(); column++) {
				columnsWidth.put(frozenAndVisibleColumns.get(column),
						widths[column]);
			}
		}

		if (config().isEnabledFrozenTable()) {
			if (columnsFrozen.isEmpty()) {
				hideFrozenTable = columnsFrozen.isEmpty();
				showFrozenTable(false);
			} else if (!columnsFrozen.isEmpty() && hideFrozenTable
					&& showFrozenTable) {
				showFrozenTable = false;
				hideFrozenTable = false;
				showFrozenTable(true);
			}
			hideFrozenTable = columnsFrozen.isEmpty();
		}

		int newTableDataSize = 0;
		for (int column : frozenAndVisibleColumns) {
			if (uidl.hasAttribute("" + column)) {
				newTableDataSize += 1;
			}
		}
		return newTableDataSize;
	}

	@Override
	protected void onLoad() {
		initContainers();

		activateMouseLeaveWindowEventHandler(true);
		activateWindowResizeEventHandler(true);
		activateTableScrollEventHandler(true);
	}

	@Override
	protected void onUnload() {
		activateMouseLeaveWindowEventHandler(false);
		activateWindowResizeEventHandler(false);

		activateTableOutEventHandler(false);
		activateFrozenTableScrollEventHandler(false);
		activateTableScrollEventHandler(false);

		unloadComponents();
	}

	public void sentMessage(final String eventType, final Object eventMessage) {
		if ("newFilter".equals(eventType) || "removeFilter".equals(eventType)
				|| "sorting".equals(eventType) || "freeze".equals(eventType)
				|| "searchColumns".equals(eventType)
				|| "pageReload".equals(eventType)) {
			firstRow = 0;
			scrollTop = 0;
			pager.reset();
		}

		// standard message elements
		int top = !config().isEnabledPageableTable() ? scrollTop : ((pager
				.getCurentPage() - 1) * config().getPageSize() * config()
				.getBodyRowHeight());
		int pageSize = !config().isEnabledPageableTable() ? ((int) (config()
				.getTableHeight() / config().getBodyRowHeight()) + 3)
				: config().getPageSize();
		request.put("left", scrollLeft);
		request.put("top", top);
		request.put("width", BrowserInfo.get().getScreenWidth());
		request.put("pageSize", pageSize);
		request.put("bodyRowHeight", config().getBodyRowHeight());

		// non-standard message elements
		request.put(eventType, eventMessage);

		// send request to server
		client.updateVariable(uidl, "request", request, true);
		request.clear();
	}

	public boolean sendingMessage() {
		return sendingMessage;
	}

	public void sendMessage(final String eventType, Object eventMessage,
			boolean showLoadingMessage) {
		disableEventsHandler();
		if (showLoadingMessage && config().isEnabledLoadingMessage()) {
			loadingMessage.show();
			loadingMessage.sendMessage(eventType, eventMessage);
		} else {
			loadingMessage.hide();
			sentMessage(eventType, eventMessage);
		}
	}

	public void updateFromUIDL(UIDL uidl, ApplicationConnection client) {
		if (client.updateComponent(this, uidl, true)) {
			return;
		}

		this.client = client;
		this.uidl = uidl.getId();
		this.request.clear();

		// START: this must be checked first
		if (uidl.getMapAttribute("translation") != null
				&& uidl.getMapAttribute("translation").getKeySet().size() > 0) {
			translation().process(uidl.getMapAttribute("translation"));
		}

		if (uidl.getMapAttribute("tableExport") != null
				&& uidl.getMapAttribute("tableExport").getKeySet().size() > 0) {
			// table configuration
			exporter().process(uidl.getMapAttribute("tableExport"));
		}

		if (uidl.getMapAttribute("tableConfig") != null
				&& uidl.getMapAttribute("tableConfig").getKeySet().size() > 0) {
			// table configuration
			config().process(uidl.getMapAttribute("tableConfig"));

			if (uidl.hasAttribute("formatedColumns")) {
				config().setColumnsFormatted(
						uidl.getIntArrayAttribute("formatedColumns"));
			}

			if (uidl.hasAttribute("enableFreezeColumns")) {
				config().enableColumnsFreeze(
						uidl.getStringArrayAttribute("enableFreezeColumns"));
			}

			if (uidl.hasAttribute("enableUnfreezeColumns")) {
				config().enableColumnsUnfreeze(
						uidl.getStringArrayAttribute("enableUnfreezeColumns"));
			}

			if (uidl.hasAttribute("enableHideColumns")) {
				config().enableColumnsHide(
						uidl.getStringArrayAttribute("enableHideColumns"));
			}

			if (config().isEnabledFrozenTable()) {
				tableFrozenWidth = Math.max(config()
						.getTableFrozenMinimalWidth(), tableFrozenWidth);
				appendFrozenTable();
				showFrozenTable = !config().isEnabledFrozenTableVisible();
				showFrozenTable(!showFrozenTable);
				activateFrozenTableScrollEventHandler(true);
				activateTableScrollEventHandler(true); // don't remove
			} else {
				tableFrozenWidth = 0;
			}

			loadComponents();
		}
		// END: this must be checked first

		if (uidl.hasAttribute("refreshVisibleData")) {
			sendMessage("newData", Boolean.FALSE.toString(),
					uidl.getBooleanAttribute("refreshVisibleData"));
		} else if (uidl.hasAttribute("emptyData")) {
			// no data to display
			getElement().setInnerHTML(
					translation().get(Translation.NO_DATA_TO_DISPLAY));
		} else if (uidl.hasAttribute("noData")) {
			if ("searchingColumnsError".equals(uidl
					.getStringAttribute("noData"))) {
				searchColumns().markSearchingField(
						uidl.getIntAttribute("searchingColumnsError"));
				// hide loading message
				loadingMessage.hide();
			} else {
				if (uidl.hasAttribute("emptyQuickSearch")) {
					quickSearch().emptyQuickSearch();
				}
				if (uidl.hasAttribute("emptySearchColumn")) {
					searchColumns().emptySearchingColumn(
							uidl.getIntAttribute("emptySearchColumn"));
				}
				sendMessage("noData", "noData", false);
			}
		} else if (uidl.hasAttribute("dataError")) {
			sendMessage("dataError", "dataError", false);
		} else {
			if (uidl.hasAttribute("allColumnsVisible")) {
				allColumnsVisible = uidl
						.getBooleanAttribute("allColumnsVisible");
				// allColumnsVisible &= enableAllColumns;
				tableHeader.set100percentWidth(allColumnsVisible);
				tableBody.set100percentWidth(allColumnsVisible);
			}

			if (uidl.hasAttribute("allRowsVisible")) {
				allRowsVisible = uidl.getBooleanAttribute("allRowsVisible");
			}

			if (uidl.hasAttribute("firstRow")) {
				firstRow = uidl.getIntAttribute("firstRow");
			}

			if (uidl.hasAttribute("allColumns")) {
				allColumns = uidl.getIntAttribute("allColumns");
				allColumns = allColumns > 0 ? allColumns : 0;
			}

			if (uidl.hasAttribute("allRows")
					&& allRows != uidl.getIntAttribute("allRows")) {
				allRows = uidl.getIntAttribute("allRows");
				pager.setAllPages(allRows);
			}

			if (uidl.hasAttribute("tableWidth")) {
				tableWidth = uidl.getIntAttribute("tableWidth");
			}

			if (uidl.hasAttribute("isAnyFrozenColumnHidden")) {
				isAnyFrozenColumnHidden = uidl
						.getBooleanAttribute("isAnyFrozenColumnHidden");
			}

			int newTableDataSize = prepareTable(uidl);
			if (newTableDataSize != (columnsFrozen.size() + columnsVisible
					.size())
					|| !uidl.hasAttribute("left")
					|| !uidl.hasAttribute("top")) {
				// frozen size + non-frozen size != data size
				sendMessage("transmissionError", "transmissionError", false);
			} else if (uidl.hasAttribute("visibleColumns")
					&& uidl.hasAttribute("left") && uidl.hasAttribute("top")) {
				left = uidl.getIntAttribute("left");
				top = config().isEnabledPageableTable() ? 0 : uidl
						.getIntAttribute("top");

				// table header
				tableHeader.readUIDL(uidl);
				tableHeader.show();

				// add/remove search columns
				searchColumns.readUIDL(uidl);
				searchColumns.show();

				// add/remove sorting
				sorting.readUIDL(uidl);
				sorting.apply();

				// add/remove filters
				filtering.readUIDL(uidl);
				filtering.apply();

				// insert data into table body
				tableBody.readUIDL(uidl);
				tableBody.show();

				// set container dimensions
				setContainerWidthAndHeight();

				// hide loading message
				loadingMessage.hide();

				// page refresh: have to be called last because
				// the table context have to be displayed first
				if (uidl.hasAttribute("pageRefresh")) {
					(new Timer() {
						public void run() {
							sendMessage("newData", Boolean.FALSE.toString(),
									false);
						}
					}).schedule(uidl.getIntAttribute("pageRefresh"));
				}
			}
		}
	}

}
