/**
 * Copyright (C) 2012  Dušan Vejnovič  <vaadin@dussan.org>
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 * http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */
package org.dussan.vaadin.dtable.properties;

import java.io.Serializable;
import java.util.ArrayList;
import java.util.Arrays;
import java.util.Collections;
import java.util.HashMap;
import java.util.List;
import java.util.Map;

import org.apache.commons.lang3.ArrayUtils;
import org.dussan.vaadin.dtable.config.formatting.Formatting;
import org.dussan.vaadin.dtable.processors.FormattingProcessor;
import org.dussan.vaadin.dtable.processors.SearchingProcessor;
import org.dussan.vaadin.dtable.properties.filtering.Filtering;
import org.dussan.vaadin.dtable.properties.filtering.FilteringHelper;
import org.dussan.vaadin.dtable.properties.searching.QuickSearchHelper;
import org.dussan.vaadin.dtable.properties.searching.SearchingColumnsHelper;
import org.dussan.vaadin.dtable.properties.searching.Searching;
import org.dussan.vaadin.dtable.properties.sorting.Sorting;
import org.dussan.vaadin.dtable.properties.sorting.SortingHelper;
import org.dussan.vaadin.dtable.properties.sorting.SortingType;

public class TableProperties implements Serializable, SortingHelper,
		FilteringHelper, QuickSearchHelper, SearchingColumnsHelper {

	private enum ColumnStatus {
		HIDDEN, FROZEN, FROZEN_AND_HIDDEN, REMOVED, UNDEFINED, UNVISIBLE, VISIBLE;
	};

	private static final long serialVersionUID = 609108827865239112L;
	private long allRows = 0;
	private int left = 0;
	private int top = 0;
	private int firstColumn = 0;
	private int firstRow = 0;
	private int tableWidth = 0;
	private int screenWidth = 0;
	private String translatedWordIsEmpty = null;
	private String translatedWordIsNotEmpty = null;

	private List<List<Object>> originTableData = new ArrayList<List<Object>>();
	private List<String> columnsHeader = new ArrayList<String>();
	private Map<Integer, String> headersTooltip = new HashMap<Integer, String>();
	private Map<Integer, ColumnStatus> columnsStatus = new HashMap<Integer, ColumnStatus>();
	private Map<Integer, List<Object>> columnsObjects = new HashMap<Integer, List<Object>>();

	private String quickSearch = new String();
	private String lastQuickSearch = new String();
	private Searching lastSearchingColumn = null;
	private Filtering lastFilteringColumn = null;

	private String getTranslatedWordIsEmpty() {
		return translatedWordIsEmpty;
	}

	public void setTranslatedWordIsEmpty(String translatedWordIsEmpty) {
		this.translatedWordIsEmpty = translatedWordIsEmpty;
	}

	private String getTranslatedWordIsNotEmpty() {
		return translatedWordIsNotEmpty;
	}

	public void setTranslatedWordIsNotEmpty(String translatedWordIsNotEmpty) {
		this.translatedWordIsNotEmpty = translatedWordIsNotEmpty;
	}

	public Object[] getColumnsDataType() {
		List<Object> dataTypes = new ArrayList<Object>();
		for (List<Object> columnData : originTableData) {
			Object type = null;
			for (Object data : columnData) {
				if (type == null) {
					type = data.getClass();
				} else {
					break;
				}
			}
			dataTypes.add(type);
		}
		return dataTypes.toArray();
	}

	public Object getOriginTableData(int column, int row) {
		return originTableData.get(column).get(row);
	}

	public List<Object> getOriginTableData(int column) {
		return originTableData.get(column);
	}

	public void resetOriginTableData() {
		originTableData.clear();
	}

	public boolean isOriginTableDataEmpty() {
		return originTableData.isEmpty();
	}

	public void setOriginTableData(List<Object> originTableData) {
		this.originTableData.add(originTableData);
	}

	public int getColumnHeaderIndex(String header) {
		try {
			return columnsHeader.indexOf(header);
		} catch (Exception e) {
			return -1;
		}
	}

	public String getColumnHeader(int col) {
		try {
			return columnsHeader.get(col);
		} catch (Exception e) {
			return null;
		}
	}

	public void setColumnHeader(int column, String header) {
		if (header != null) {
			columnsHeader.add(column, header);
			if (!columnsStatus.containsKey(column)) {
				setColumnStatus(column, ColumnStatus.UNVISIBLE);
			}
		}
	}

	public String[] getCompleteColumnsHeader() {
		try {
			return columnsHeader.toArray(new String[] {});
		} catch (Exception e) {
			return null;
		}
	}

	public String[] getColumnsHeader(int[] columns) {
		try {
			List<String> headers = new ArrayList<String>();
			for (int column : columns) {
				if (!isRemovedColumn(column) && !isHiddenColumn(column)) {
					headers.add(columnsHeader.get(column));
				}
			}
			return headers.toArray(new String[] {});
		} catch (Exception e) {
			return null;
		}
	}

	public String[] getColumnsHeader() {
		try {
			List<String> headers = new ArrayList<String>();
			for (int column = 0; column < columnsHeader.size(); column++) {
				if (!isRemovedColumn(column) && !isHiddenColumn(column)) {
					headers.add(columnsHeader.get(column));
				}
			}
			return headers.toArray(new String[] {});
		} catch (Exception e) {
			return null;
		}
	}

	public void setColumnsHeader(String... headers) {
		if (headers != null && headers.length > 0) {
			int column = columnsStatus.size();
			for (String header : headers) {
				columnsHeader.add(header);
				setColumnStatus(column++, ColumnStatus.UNVISIBLE);
			}
		}
	}

	public String getHeaderTooltip(int col) {
		try {
			if (headersTooltip.containsKey(col)) {
				return headersTooltip.get(col);
			} else {
				return "";
			}
		} catch (Exception e) {
			return "";
		}
	}

	public void setHeaderTooltip(int column, String tooltip) {
		if (tooltip != null) {
			headersTooltip.put(column, tooltip);
		}
	}

	public String[] getHeadersTooltip(int[] columns) {
		try {
			List<String> tooltips = new ArrayList<String>();
			for (int column : columns) {
				if (!isRemovedColumn(column) && !isHiddenColumn(column)) {
					tooltips.add(getHeaderTooltip(column));
				}
			}
			return tooltips.toArray(new String[] {});
		} catch (Exception e) {
			return null;
		}
	}

	private boolean hasAnyColumnStatus(ColumnStatus status) {
		return columnsStatus.containsValue(status);
	}

	private ColumnStatus getColumnStatus(int column) {
		if (columnsStatus.containsKey(column)) {
			return columnsStatus.get(column);
		}
		return ColumnStatus.UNDEFINED;
	}

	private void setColumnStatus(int column, ColumnStatus columnStatus) {
		columnsStatus.put(column, columnStatus);
	}

	private boolean hasAnyColumnsObjects() {
		return !columnsObjects.isEmpty();
	}

	private void removeColumnObjects(int column) {
		if (hasAnyColumnsObjects()) {
			if (hasColumnObjects(column)) {
				columnsObjects.remove(column);
			}
		}
	}

	private List<Integer> getColumnsObjectsKeys() {
		if (hasAnyColumnsObjects()) {
			List<Integer> columns = new ArrayList<Integer>(
					columnsObjects.keySet());
			Collections.sort(columns);
			return columns;
		}
		return new ArrayList<Integer>();
	}

	private boolean hasColumnObjects(int column) {
		if (hasAnyColumnsObjects()) {
			return columnsObjects.containsKey(column);
		}
		return false;
	}

	private List<Object> getColumnObjects(int column) {
		if (hasAnyColumnsObjects()) {
			if (hasColumnObjects(column)) {
				return columnsObjects.get(column);
			}
		}
		return new ArrayList<Object>();
	}

	private void setColumnObjects(int column, List<Object> columnObject) {
		if (columnObject != null && !columnObject.isEmpty()) {
			columnsObjects.put(column, columnObject);
		} else if (hasAnyColumnsObjects()) {
			columnsObjects.remove(column);
		}
	}

	private void resetVisibleColumns() {
		for (int column = 0; column < getAllColumns(); column++) {
			switch (getColumnStatus(column)) {
			case VISIBLE:
				setColumnStatus(column, ColumnStatus.UNVISIBLE);
				break;
			default:
				break;
			}
		}
	}

	private int getVisibleColumnsSize() {
		int size = 0;
		for (int column = 0; column < getAllColumns(); column++) {
			switch (getColumnStatus(column)) {
			case VISIBLE:
				size += 1;
				break;
			default:
				break;
			}
		}
		return size;
	}

	public void checkVisibleColumns(int tableWidth, int left, int top,
			int bodyRowHeight, int[] columnWidths, boolean enableAllColumns) {
		this.screenWidth = tableWidth;
		this.left = left;
		this.top = top;
		calculateLeftAndTopValues(bodyRowHeight, columnWidths);
		checkVisibleColumns(tableWidth, columnWidths, enableAllColumns);
	}

	private void calculateLeftAndTopValues(int bodyRowHeight, int[] columnWidths) {
		int width = 0;
		for (int column = 0; column < getAllColumns(); column++) {
			if (!isRemovedColumn(column) && !isHiddenColumn(column)
					&& !isFrozenColumn(column)) {
				if ((width + columnWidths[column]) > left) {
					firstColumn = column;
					left = width;
					break;
				}
				width += columnWidths[column];
			}
		}

		firstRow = top / bodyRowHeight;
		top = firstRow * bodyRowHeight;
	}

	public void checkVisibleColumns(int tableWidth, int[] columnWidths,
			boolean enableAllColumns) {
		if (columnWidths != null && columnWidths.length > 0) {
			resetVisibleColumns();
			firstColumn = enableAllColumns ? 0 : firstColumn;

			int width = 0;
			tableWidth += columnWidths[firstColumn];
			for (int column = firstColumn; column < getAllColumns(); column++) {
				if (!isRemovedColumn(column) && !isHiddenColumn(column)
						&& !isFrozenColumn(column)) {
					if (!enableAllColumns) {
						setColumnStatus(column, ColumnStatus.VISIBLE);
						if ((width += columnWidths[column]) > tableWidth) {
							break;
						}
					} else {
						setColumnStatus(column, ColumnStatus.VISIBLE);
					}
				}
			}

			if (!enableAllColumns && width < tableWidth) {
				tableWidth -= columnWidths[firstColumn];
				for (int column = firstColumn - 1; 0 <= column; column--) {
					if (!isRemovedColumn(column) && !isHiddenColumn(column)
							&& !isFrozenColumn(column)) {
						setColumnStatus(column, ColumnStatus.VISIBLE);
						width += columnWidths[column];
						left -= columnWidths[column];
						if (width > tableWidth) {
							firstColumn = column;
							break;
						}
					}
				}
			}
		}
	}

	public void calculateTableWidth(int[] columnWidths) {
		tableWidth = 0;
		for (int column = 0; column < getAllColumns(); column++) {
			if (!isRemovedColumn(column) && !isHiddenColumn(column)
					&& !isFrozenColumn(column)) {
				tableWidth += columnWidths[column];
			}
		}
	}

	public int getTableWidth() {
		return tableWidth;
	}

	public int getScreenWidth() {
		return screenWidth;
	}

	public int getLeft() {
		return left;
	}

	public int getTop() {
		return top;
	}

	public int getFirstRow() {
		return firstRow;
	}

	public void setFirstRow(int firstRow) {
		this.firstRow = firstRow;
	}

	public boolean isAllColumnsVisible() {
		return (getAllColumns() == (getRemovedColumnsSize()
				+ getFrozenColumnsSize() + getHiddenColumnsSize() + getVisibleColumnsSize()));
	}

	public int getAllColumns() {
		return columnsHeader.size();
	}

	public long getAllRows() {
		return allRows;
	}

	public void setAllRows(long allRows) {
		this.allRows = allRows >= 0 ? allRows : 0;
	}

	public int[] getFrozenAndVisibleColumns() {
		List<Integer> columns = new ArrayList<Integer>();
		for (int column = 0; column < getAllColumns(); column++) {
			switch (getColumnStatus(column)) {
			case FROZEN:
			case VISIBLE:
				columns.add(column);
				break;
			default:
				break;
			}
		}
		return ArrayUtils.toPrimitive(columns.toArray(new Integer[] {}));
	}

	public int[] getFrozenAndAllVisibleColumns() {
		List<Integer> columns = new ArrayList<Integer>();
		for (int column = 0; column < getAllColumns(); column++) {
			switch (getColumnStatus(column)) {
			case FROZEN:
			case UNVISIBLE:
			case VISIBLE:
				columns.add(column);
				break;
			default:
				break;
			}
		}
		return ArrayUtils.toPrimitive(columns.toArray(new Integer[] {}));
	}

	public int[] getVisibleColumns() {
		List<Integer> columns = new ArrayList<Integer>();
		for (int column = 0; column < getAllColumns(); column++) {
			switch (getColumnStatus(column)) {
			case VISIBLE:
				columns.add(column);
				break;
			default:
				break;
			}
		}
		return ArrayUtils.toPrimitive(columns.toArray(new Integer[] {}));
	}

	public void resetRemovedColumns() {
		for (int column = 0; column < getAllColumns(); column++) {
			switch (getColumnStatus(column)) {
			case REMOVED:
				setColumnStatus(column, ColumnStatus.UNVISIBLE);
				break;
			default:
				break;
			}
		}
	}

	public boolean isNoRemovedColumn() {
		return !hasAnyColumnStatus(ColumnStatus.REMOVED);
	}

	public boolean isRemovedColumn(int column) {
		switch (getColumnStatus(column)) {
		case REMOVED:
			return true;
		default:
			return false;
		}
	}

	private int getRemovedColumnsSize() {
		int size = 0;
		for (int column = 0; column < getAllColumns(); column++) {
			switch (getColumnStatus(column)) {
			case REMOVED:
				size += 1;
				break;
			default:
				break;
			}
		}
		return size;
	}

	public int[] getRemovedColumns() {
		List<Integer> columns = new ArrayList<Integer>();
		for (int column = 0; column < getAllColumns(); column++) {
			switch (getColumnStatus(column)) {
			case REMOVED:
				columns.add(column);
				break;
			default:
				break;
			}
		}
		return ArrayUtils.toPrimitive(columns.toArray(new Integer[] {}));
	}

	public void setRemovedColumn(int column) {
		setColumnStatus(column, ColumnStatus.REMOVED);
		removeColumnObjects(column);
	}

	public void setRemovedColumns(int... columns) {
		for (int column : columns) {
			setRemovedColumn(column);
		}
	}

	public void resetHiddenColumns() {
		for (int column = 0; column < getAllColumns(); column++) {
			switch (getColumnStatus(column)) {
			case FROZEN_AND_HIDDEN:
				setColumnStatus(column, ColumnStatus.FROZEN);
				break;
			case HIDDEN:
				setColumnStatus(column, ColumnStatus.UNVISIBLE);
				break;
			default:
				break;
			}
		}
	}

	public boolean isNoHiddenColumn() {
		return !(hasAnyColumnStatus(ColumnStatus.FROZEN_AND_HIDDEN) || hasAnyColumnStatus(ColumnStatus.HIDDEN));
	}

	public boolean isHiddenColumn(int column) {
		switch (getColumnStatus(column)) {
		case FROZEN_AND_HIDDEN:
		case HIDDEN:
			return true;
		default:
			return false;
		}
	}

	private int getHiddenColumnsSize() {
		int size = 0;
		for (int column = 0; column < getAllColumns(); column++) {
			switch (getColumnStatus(column)) {
			case FROZEN_AND_HIDDEN:
			case HIDDEN:
				size += 1;
				break;
			default:
				break;
			}
		}
		return size;
	}

	public void removeHiddenColumn(int column) {
		switch (getColumnStatus(column)) {
		case FROZEN_AND_HIDDEN:
			setColumnStatus(column, ColumnStatus.FROZEN);
			break;
		case HIDDEN:
			setColumnStatus(column, ColumnStatus.UNVISIBLE);
			break;
		default:
			break;
		}
	}

	public boolean isHiddenColumnFrozen(int column) {
		switch (getColumnStatus(column)) {
		case FROZEN_AND_HIDDEN:
			return true;
		default:
			return false;
		}
	}

	public void setHiddenColumn(int column) {
		switch (getColumnStatus(column)) {
		case FROZEN:
			setColumnStatus(column, ColumnStatus.FROZEN_AND_HIDDEN);
			break;
		default:
			setColumnStatus(column, ColumnStatus.HIDDEN);
			break;
		}
	}

	public int[] getHiddenColumns() {
		List<Integer> columns = new ArrayList<Integer>();
		for (int column = 0; column < getAllColumns(); column++) {
			switch (getColumnStatus(column)) {
			case FROZEN_AND_HIDDEN:
			case HIDDEN:
				columns.add(column);
				break;
			default:
				break;
			}
		}
		return ArrayUtils.toPrimitive(columns.toArray(new Integer[] {}));
	}

	public String[] getHiddenColumnsString() {
		List<String> columns = new ArrayList<String>();
		for (int column = 0; column < getAllColumns(); column++) {
			switch (getColumnStatus(column)) {
			case FROZEN_AND_HIDDEN:
				columns.add("-" + column);
				break;
			case HIDDEN:
				columns.add("+" + column);
				break;
			default:
				break;
			}
		}
		return columns.toArray(new String[] {});
	}

	public void setHiddenColumns(int... columns) {
		for (int column : columns) {
			setHiddenColumn(column);
		}
	}

	public void resetFrozenColumns() {
		for (int column = 0; column < getAllColumns(); column++) {
			switch (getColumnStatus(column)) {
			case FROZEN_AND_HIDDEN:
				setColumnStatus(column, ColumnStatus.HIDDEN);
				break;
			case FROZEN:
				setColumnStatus(column, ColumnStatus.UNVISIBLE);
				break;
			default:
				break;
			}
		}
	}

	public boolean isNoFrozenColumn() {
		return !hasAnyColumnStatus(ColumnStatus.FROZEN);
	}

	public boolean isFrozenColumn(int column) {
		switch (getColumnStatus(column)) {
		case FROZEN:
			return true;
		default:
			return false;
		}
	}

	private int getFrozenColumnsSize() {
		int size = 0;
		for (int column = 0; column < getAllColumns(); column++) {
			switch (getColumnStatus(column)) {
			case FROZEN:
				size += 1;
				break;
			default:
				break;
			}
		}
		return size;
	}

	public boolean isAnyFrozenColumnHidden() {
		return hasAnyColumnStatus(ColumnStatus.FROZEN_AND_HIDDEN);
	}

	public void removeFrozenColumn(int column) {
		switch (getColumnStatus(column)) {
		case FROZEN:
			setColumnStatus(column, ColumnStatus.UNVISIBLE);
			break;
		default:
			break;
		}
	}

	public void setFrozenColumn(int column) {
		switch (getColumnStatus(column)) {
		case HIDDEN:
			setColumnStatus(column, ColumnStatus.FROZEN_AND_HIDDEN);
			break;
		default:
			setColumnStatus(column, ColumnStatus.FROZEN);
			break;
		}
	}

	public int[] getFrozenColumns() {
		List<Integer> columns = new ArrayList<Integer>();
		for (int column = 0; column < getAllColumns(); column++) {
			switch (getColumnStatus(column)) {
			case FROZEN:
				columns.add(column);
				break;
			default:
				break;
			}
		}
		return ArrayUtils.toPrimitive(columns.toArray(new Integer[] {}));
	}

	public void setFrozenColumns(int... columns) {
		for (int column : columns) {
			setFrozenColumn(column);
		}
	}

	private void replaceSortingColumn(Sorting sorting) {
		int column = sorting.getColumn();
		if (isSortingColumn(column)) {
			removeSortingColumn(column);
		}
		List<Object> objects = getColumnObjects(column);
		objects.add(sorting);
		setColumnObjects(column, objects);
	}

	@Override
	public void resetSortingColumns() {
		if (hasAnyColumnsObjects()) {
			for (int column : getColumnsObjectsKeys()) {
				if (isSortingColumn(column)) {
					removeSortingColumn(column);
				}
			}
		}
	}

	@Override
	public boolean isNoSortingColumn() {
		if (hasAnyColumnsObjects()) {
			for (Integer column : getColumnsObjectsKeys()) {
				if (hasColumnObjects(column)) {
					for (Object object : getColumnObjects(column)) {
						if (object instanceof Sorting) {
							return false;
						}
					}
				}
			}
		}
		return true;
	}

	@Override
	public void removeSortingColumn(int column) {
		if (hasAnyColumnsObjects()) {
			if (isSortingColumn(column)) {
				List<Object> objects = getColumnObjects(column);
				for (int object = objects.size() - 1; 0 <= object; object--) {
					if (objects.get(object) instanceof Sorting) {
						objects.remove(object);
						break;
					}
				}
				setColumnObjects(column, objects);
			}
		}
	}

	@Override
	public boolean isSortingColumn(int column) {
		if (hasAnyColumnsObjects()) {
			if (!isNoSortingColumn() && hasColumnObjects(column)) {
				for (Object object : getColumnObjects(column)) {
					if (object instanceof Sorting) {
						return true;
					}
				}
			}
		}
		return false;
	}

	@Override
	public Sorting getSortingColumn(int column) {
		if (!isNoSortingColumn() && isSortingColumn(column)) {
			for (Object object : getColumnObjects(column)) {
				if (object instanceof Sorting) {
					return (Sorting) object;
				}
			}
		}
		return null;
	}

	@Override
	public void setSortingColumn(int column, SortingType type) {
		if (type != null) {
			replaceSortingColumn(new Sorting(column, null, type));
		}
	}

	@Override
	public Sorting[] getSortingColumns() {
		List<Sorting> sortings = new ArrayList<Sorting>();
		if (!isNoSortingColumn()) {
			for (Integer column : getColumnsObjectsKeys()) {
				if (isSortingColumn(column)) {
					sortings.add(getSortingColumn(column));
				}
			}
		}
		return sortings.toArray(new Sorting[] {});
	}

	private String[] getSortingColumnsString() {
		List<String> sortingString = new ArrayList<String>();
		if (!isNoSortingColumn()) {
			for (Integer column : getColumnsObjectsKeys()) {
				if (isSortingColumn(column)) {
					sortingString.add(getSortingColumn(column).toString());
				}
			}
			return sortingString.toArray(new String[] {});
		}
		return (new String[] { "none" });
	}

	@Override
	public void setSortingColumns(SortingType type, int... columns) {
		if (type != null && columns != null && columns.length > 0) {
			for (int column : columns) {
				replaceSortingColumn(new Sorting(column, null, type));
			}
		}
	}

	@Override
	public void setSortingColumns(Sorting[] sortings) {
		if (sortings != null && sortings.length > 0) {
			for (Sorting sorting : sortings) {
				replaceSortingColumn(sorting);
			}
		}
	}

	private void newFilteringColumn(Filtering filtering) {
		int column = filtering.getColumn();
		if (isFilteringColumn(column)) {
			lastFilteringColumn = getFilteringColumn(column);
			removeFilteringColumn(column);
		} else {
			lastFilteringColumn = new Filtering(column, null, null, null, null);
		}
		List<Object> objects = getColumnObjects(column);
		objects.add(filtering);
		setColumnObjects(column, objects);
	}

	@Override
	public void removeFilteringColumn(int column) {
		if (hasAnyColumnsObjects()) {
			if (isFilteringColumn(column)) {
				List<Object> objects = getColumnObjects(column);
				for (int object = objects.size() - 1; 0 <= object; object--) {
					if (objects.get(object) instanceof Filtering
							&& ((Filtering) objects.get(object)).getColumn() >= 0) {
						objects.remove(object);
						break;
					}
				}
				setColumnObjects(column, objects);
			}
		}
	}

	@Override
	public void resetFilteringColumns(String newDataType) {
		if ("newFiltering".equals(newDataType)) {
			int column = lastFilteringColumn.getColumn();
			removeFilteringColumn(column);
			if (lastFilteringColumn.getHeader() != null) {
				List<Object> objects = getColumnObjects(column);
				objects.add(lastFilteringColumn);
				setColumnObjects(column, objects);
			}
			lastFilteringColumn = null;
		}
	}

	@Override
	public boolean isNoFilteringColumn() {
		if (hasAnyColumnsObjects()) {
			for (Integer column : getColumnsObjectsKeys()) {
				if (hasColumnObjects(column)) {
					for (Object object : getColumnObjects(column)) {
						if (object instanceof Filtering
								&& ((Filtering) object).getColumn() >= 0) {
							return false;
						}
					}
				}
			}
		}
		return true;
	}

	@Override
	public boolean isFilteringColumn(int column) {
		if (!isNoFilteringColumn() && hasColumnObjects(column)) {
			for (Object object : getColumnObjects(column)) {
				if (object instanceof Filtering
						&& ((Filtering) object).getColumn() >= 0) {
					return true;
				}
			}
		}
		return false;
	}

	@Override
	public Filtering getFilteringColumn(int column) {
		if (!isNoFilteringColumn() && isFilteringColumn(column)) {
			for (Object object : getColumnObjects(column)) {
				if (object instanceof Filtering
						&& ((Filtering) object).getColumn() >= 0) {
					return (Filtering) object;
				}
			}
		}
		return null;
	}

	@Override
	public Filtering[] getFilteringColumns() {
		List<Filtering> filtering = new ArrayList<Filtering>();
		if (!isNoFilteringColumn()) {
			for (Integer column : getColumnsObjectsKeys()) {
				if (isFilteringColumn(column)) {
					filtering.add(getFilteringColumn(column));
				}
			}
		}
		return filtering.toArray(new Filtering[] {});
	}

	@Override
	public void setFilteringColumn(Filtering newFiltering) {
		if (newFiltering != null) {
			newFilteringColumn(newFiltering);
		}
	}

	@Override
	public void resetQuickSearch(String newDataObject) {
		if ("quickSearch".equals(newDataObject)) {
			quickSearch = lastQuickSearch;
			lastQuickSearch = new String();
		}
	}

	@Override
	public Filtering[] getQuickSearchFormatting(Formatting[] formatting,
			boolean enableFormatting) {
		if (quickSearch != null && !quickSearch.isEmpty()) {
			List<Filtering> filterings = new ArrayList<Filtering>();
			for (int column = 0; column < getAllColumns(); column++) {
				String quickSearch = this.quickSearch;
				if (enableFormatting) {
					for (Formatting format : formatting) {
						if (format.getColumn().equals(column)) {
							Object search = FormattingProcessor.partialProcess(
									format, this.quickSearch);
							quickSearch = search == null ? this.quickSearch
									: search.toString();
							break;
						}
					}
				}
				quickSearch = "%" + quickSearch + "%";
				filterings.add(new Filtering(column, getColumnHeader(column),
						new String[] { "LIKE" }, new String[] { "%"
								+ this.quickSearch + "%" },
						new String[] { quickSearch }));
			}
			return filterings.toArray(new Filtering[] {});
		}
		return null;
	}

	@Override
	public String getQuickSearch() {
		return quickSearch;
	}

	@Override
	public void setQuickSearch(String quickSearch) {
		this.lastQuickSearch = this.quickSearch;
		this.quickSearch = quickSearch != null ? quickSearch : this.quickSearch;
	}

	private void newSearchingColumn(Searching searchColumn) {
		int column = searchColumn.getColumn();
		if (isSearchingColumn(column)) {
			lastSearchingColumn = getSearchingColumn(column);
			removeSearchingColumn(column);
		} else {
			lastSearchingColumn = new Searching(column, null, null, null, null);
		}
		List<Object> objects = getColumnObjects(column);
		objects.add(searchColumn);
		setColumnObjects(column, objects);
	}

	@Override
	public void removeSearchingColumns() {
		if (!isNoSearchingColumn()) {
			for (Searching search : getSearchingColumns()) {
				removeSearchingColumn(search.getColumn());
			}
		}
	}

	@Override
	public void removeSearchingColumn(int column) {
		if (hasAnyColumnsObjects()) {
			if (isSearchingColumn(column)) {
				List<Object> objects = getColumnObjects(column);
				for (int object = objects.size() - 1; 0 <= object; object--) {
					if (objects.get(object) instanceof Searching) {
						objects.remove(object);
						break;
					}
				}
				setColumnObjects(column, objects);
			}
		}
	}

	@Override
	public void resetSearchingColumns(String newDataType) {
		if ("searchingColumns".equals(newDataType)) {
			int column = lastSearchingColumn.getColumn();
			removeSearchingColumn(column);
			if (lastSearchingColumn.getHeader() != null) {
				List<Object> objects = getColumnObjects(column);
				objects.add(lastSearchingColumn);
				setColumnObjects(column, objects);
			}
			lastSearchingColumn = null;
		}
	}

	@Override
	public boolean isNoSearchingColumn() {
		if (hasAnyColumnsObjects()) {
			for (Integer column : getColumnsObjectsKeys()) {
				if (hasColumnObjects(column)) {
					for (Object object : getColumnObjects(column)) {
						if (object instanceof Searching) {
							return false;
						}
					}
				}
			}
		}
		return true;
	}

	@Override
	public boolean isSearchingColumn(int column) {
		if (!isNoSearchingColumn() && hasColumnObjects(column)) {
			for (Object object : getColumnObjects(column)) {
				if (object instanceof Searching) {
					return true;
				}
			}
		}
		return false;
	}

	@Override
	public Searching getSearchingColumn(int column) {
		if (!isNoSearchingColumn() && isSearchingColumn(column)) {
			for (Object object : getColumnObjects(column)) {
				if (object instanceof Searching) {
					return (Searching) object;
				}
			}
		}
		return null;
	}

	@Override
	public Searching[] getSearchingColumns() {
		List<Searching> searching = new ArrayList<Searching>();
		if (!isNoSearchingColumn()) {
			for (Integer column : getColumnsObjectsKeys()) {
				if (isSearchingColumn(column)) {
					searching.add(getSearchingColumn(column));
				}
			}
		}
		return searching.toArray(new Searching[] {});
	}

	@Override
	public String[] getSearchingColumnsString() {
		if (!isNoSearchingColumn()) {
			List<String> searchString = new ArrayList<String>();
			for (Searching search : getSearchingColumns()) {
				searchString.add("" + search.getColumn());
				String operator = search.getOperator();
				if (operator.contains("LIKE")) {
					operator = operator.startsWith("LIKE") ? "LIKE"
							: "NOT LIKE";
					searchString.add(operator + " '" + search.getValue() + "'");
				} else {
					searchString.add(operator + " " + search.getValue());
				}
			}
			return new ArrayList<String>(searchString).toArray(new String[] {});
		}
		return null;
	}

	@Override
	public void setSearchingColumns(Searching newSearch) {
		if (newSearch != null) {
			newSearchingColumn(newSearch);
		}
	}

	public Filtering[] getCompleteFiltering() {
		if (!isNoFilteringColumn() || !isNoSearchingColumn()) {
			List<Filtering> filterings = new ArrayList<Filtering>();
			if (!isNoFilteringColumn()) {
				filterings.addAll(Arrays.asList(getFilteringColumns()));
			}

			if (!isNoSearchingColumn()) {
				if (filterings.isEmpty()) {
					for (Searching searching : getSearchingColumns()) {
						filterings.add(SearchingProcessor
								.searching2Filtering(searching));
					}
				} else {
					List<Filtering> tempFilterings = new ArrayList<Filtering>();
					for (Searching searching : getSearchingColumns()) {
						boolean filteringExist = false;
						for (Filtering filtering : getFilteringColumns()) {
							if (searching.getColumn().equals(
									filtering.getColumn())) {
								List<String> operators = new ArrayList<String>(
										Arrays.asList(filtering.getOperators()));
								List<String> values = new ArrayList<String>(
										Arrays.asList(filtering.getValues()));
								List<String> originalValues = new ArrayList<String>(
										Arrays.asList(filtering
												.getOriginalValues()));
								operators.add(searching.getOperator());
								values.add(searching.getValue());
								originalValues
										.add(searching.getOriginalValue());
								Filtering filter = new Filtering(
										filtering.getColumn(),
										filtering.getHeader(),
										operators.toArray(new String[] {}),
										values.toArray(new String[] {}),
										originalValues.toArray(new String[] {}));
								tempFilterings.add(filter);
								filteringExist = true;
								break;
							}
						}
						if (!filteringExist) {
							tempFilterings.add(SearchingProcessor
									.searching2Filtering(searching));
						}
					}
					for (Filtering filtering : filterings) {
						boolean filteringExist = false;
						for (Filtering filter : tempFilterings) {
							if (filtering.getColumn()
									.equals(filter.getColumn())) {
								filteringExist = true;
								break;
							}
						}
						if (!filteringExist) {
							tempFilterings.add(filtering);
						}
					}
					filterings = new ArrayList<Filtering>(tempFilterings);
				}
			}
			return filterings.toArray(new Filtering[] {});
		}
		return null;
	}

	public String[] getNewFiltering() {
		Filtering[] filtering = getCompleteFiltering();
		if (filtering != null) {
			List<String> newFiltering = new ArrayList<String>();
			for (Filtering filter : filtering) {
				for (int element = 0; element < filter.getOperators().length; element++) {
					String operator = filter.getOperators()[element];
					newFiltering.add("" + filter.getColumn());
					if (operator.contains("EMPTY")) {
						newFiltering.add(operator);
						if ("IS EMPTY".equals(operator)) {
							newFiltering.add(getTranslatedWordIsEmpty());
						} else {
							newFiltering.add(getTranslatedWordIsNotEmpty());
						}
					} else {
						String value = filter.getValues()[element];
						StringBuilder value1 = new StringBuilder("_");
						StringBuilder value2 = new StringBuilder(value);
						if (operator.endsWith("LIKE")) {
							value1.insert(0, value.startsWith("%") ? "%" : "");
							value1.append(value.endsWith("%") ? "%" : "");
							value1.insert(0, "'").append("'");
							value2.insert(0, "'").append("'");
						}
						value1.insert(0, operator + " ");
						value2.insert(0, operator + " ");
						newFiltering.add(value1.toString());
						newFiltering.add(value2.toString());
					}
				}
			}
			return newFiltering.toArray(new String[] {});
		}
		return null;
	}

	public String getFilteringHtml() {
		Filtering[] filtering = getCompleteFiltering();
		if (filtering != null) {
			StringBuilder filterString = new StringBuilder();
			for (Filtering filter : filtering) {
				filterString.append("<b>["
						+ getColumnHeader(filter.getColumn()) + "]:</b> ");
				StringBuilder operatorString = new StringBuilder();
				for (int operator = 0; operator < filter.getOperators().length; operator++) {
					operatorString.append(", ");
					if ("IS EMPTY".equals(filter.getOperators()[operator])) {
						operatorString.append(getTranslatedWordIsEmpty());
					} else if ("IS NOT EMPTY"
							.equals(filter.getOperators()[operator])) {
						operatorString.append(getTranslatedWordIsNotEmpty());
					} else {
						operatorString.append(filter.getOperators()[operator]);
						operatorString.append(" ");
						operatorString.append(filter.getOperators()[operator]
								.endsWith("LIKE") ? "'" : "");
						operatorString.append(filter.getValues()[operator]);
						operatorString.append(filter.getOperators()[operator]
								.endsWith("LIKE") ? "'" : "");
					}
				}
				filterString.append(operatorString.substring(2) + "; ");
			}
			return filterString.toString().trim();
		}
		return null;
	}

	public Map<String, Object> getPaintObjects() {
		Map<String, Object> paintObjects = new HashMap<String, Object>();

		paintObjects.put("left", getLeft());
		paintObjects.put("top", getTop());
		paintObjects.put("firstRow", getFirstRow());
		paintObjects.put("allColumns", getAllColumns());
		paintObjects.put("allRows", getAllRows());
		paintObjects.put("visibleColumns", getVisibleColumns());
		paintObjects.put("allColumnsVisible", isAllColumnsVisible());
		paintObjects.put("tableWidth", getTableWidth());
		paintObjects.put("isAnyFrozenColumnHidden", isAnyFrozenColumnHidden());

		if (!isNoSortingColumn()) {
			paintObjects.put("sorting", getSortingColumnsString());
		} else {
			paintObjects.remove("sorting");
		}

		if (!isNoFrozenColumn()) {
			paintObjects.put("frozenColumns", getFrozenColumns());
		} else {
			paintObjects.remove("frozenColumns");
		}

		return paintObjects;
	}

}
